#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <pthread.h>
#include <sys/time.h>

enum color {
    BLACK = 0,
    RED,
    GREEN,
    YELLOW,
    BLUE,
    MAGENTA,
    CYAN,
    WHITE
};

/* a drawable object */
struct object {
	int old_x;
	int old_y;
	int cur_x;
	int cur_y;
    enum color bg_color;
    enum color text_color;
	char symbol[3];
};

struct player {
    int max_bombs;
    int bombs;
    int range;
    struct object the_object;
};

struct enemy {
    int alive;
    struct object the_object;
};

struct bomb {
    int countdown;
    int range;
    struct object the_object;
};

struct fire {
    int direction;
    int lifetime;
    int delay;
    struct object the_object;
};

#define MAX_ENEMIES 32
#define MAX_BOMBS 32
#define MAX_FIRES 256
#define WORLD_WIDTH 30
#define WORLD_HEIGHT 24
#define BOMB_COUNTDOWN 50

/* the world */
int world[WORLD_WIDTH][WORLD_HEIGHT];
struct player the_player;
struct enemy the_enemies[MAX_ENEMIES];
struct bomb the_bombs[MAX_BOMBS];
struct fire the_fires[MAX_FIRES];

int lives = 3;
int level = 1;
int all_levels_completed = 0;

/* input thread */
pthread_t thread;
pthread_mutex_t button_pressed_mutex;
int button_pressed = -1;


int load_level(unsigned int level);
void init_player(int x, int y);
void add_enemy(int x, int y);
void add_bomb(int x, int y);
void add_fire(int x, int y, int lifetime, int delay, int direction);

void player_gets_hit();
void enemy_gets_hit(int x, int y);
void block_gets_hit(int x, int y);

void move_player();
void move_enemies();
void move_bombs();
void move_fires();

int getch();

void draw_level();
void draw_screen();
void draw_status();
void clear_field(int x, int y);
void draw_object(struct object* the_object);
void clear_screen();
void set_cursor(int x, int y);
void set_text_color(enum color the_color);
void set_bg_color(enum color the_color);
void reset_colors();

void* thread_main(void* arg)
{
    while (1) {
        int tmp = getch();
        
        pthread_mutex_lock(&button_pressed_mutex);
        button_pressed = tmp;
        pthread_mutex_unlock(&button_pressed_mutex);
        
        if (tmp == 'q')
            break;
    }
    
    return 0;
}

int main(int argc, char** argv)
{
    srand(time(0));
    if(-1 == load_level(level))
        all_levels_completed = 1;
    
    pthread_mutex_init(&button_pressed_mutex,0);
    pthread_create(&thread, 0, thread_main, 0);
    
    int current_level = level;
    
    int timer = 0;
	while(1) {
        usleep(50000);
        
        if (level != current_level) {
            current_level = level;
            if(-1 == load_level(level)) {
                all_levels_completed = 1;
                set_cursor(28,1);
                set_bg_color(CYAN);
                set_text_color(WHITE);
                printf("You Win");
            }
        }
        
        pthread_mutex_lock(&button_pressed_mutex);
        int input = button_pressed;
        button_pressed = -1;
        pthread_mutex_unlock(&button_pressed_mutex);
        
        if (input == 'q') {
            exit(0);
        } else if (input == 'n') {
            lives = 3;
            level = 1;
            if(-1 == load_level(level))
                exit(1);
        } else {
            move_player(input);
        }
        
        move_bombs();
        move_fires();
        
        if (timer > 10) {
            move_enemies();
            timer = 0;
        }
        timer++;
	}
    
    pthread_join(thread,0);
    
	return 0;
}

void init_player(int x, int y)
{
    the_player.the_object.old_x = x;
	the_player.the_object.old_y = y;
	the_player.the_object.cur_x = x;
	the_player.the_object.cur_y = y;
    the_player.the_object.bg_color = GREEN;
    the_player.the_object.text_color = WHITE;
	strcpy(the_player.the_object.symbol,"  ");
    
    the_player.max_bombs = 1;
    the_player.bombs = 0;
    the_player.range = 1;
}

void add_enemy(int x, int y)
{
    int i;
    for (i = 0; i < MAX_ENEMIES; i++)
    {
        if (the_enemies[i].alive == 0)
        {
            the_enemies[i].the_object.old_x = x;
            the_enemies[i].the_object.old_y = y;
            the_enemies[i].the_object.cur_x = x;
            the_enemies[i].the_object.cur_y = y;
            the_enemies[i].the_object.bg_color = RED;
            the_enemies[i].the_object.text_color = WHITE;
            strcpy(the_enemies[i].the_object.symbol,"  ");
            
            the_enemies[i].alive = 1;
            
            return;
        }
    }
}

void add_bomb(int x, int y)
{
    int i;
    for (i = 0; i < MAX_BOMBS; i++)
    {
        if (the_bombs[i].countdown > 0)
        {
            if (the_bombs[i].the_object.cur_x == x && the_bombs[i].the_object.cur_y == y)
                return;
        }
    }
    
    
    for (i = 0; i < MAX_BOMBS; i++)
    {
        if (the_bombs[i].countdown == 0)
        {
            the_bombs[i].the_object.old_x = x;
            the_bombs[i].the_object.old_y = y;
            the_bombs[i].the_object.cur_x = x;
            the_bombs[i].the_object.cur_y = y;
            the_bombs[i].the_object.bg_color = MAGENTA;
            the_bombs[i].the_object.text_color = WHITE;
            strcpy(the_bombs[i].the_object.symbol,"  ");
            
            the_bombs[i].countdown = BOMB_COUNTDOWN;
            the_bombs[i].range = the_player.range;
            
            the_player.bombs++;
            
            return;
        }
    }
}

void add_fire(int x, int y, int lifetime, int delay, int direction)
{
    int i;
    for (i = 0; i < MAX_FIRES; i++)
    {
        if (the_fires[i].lifetime == 0)
        {
            the_fires[i].the_object.old_x = x;
            the_fires[i].the_object.old_y = y;
            the_fires[i].the_object.cur_x = x;
            the_fires[i].the_object.cur_y = y;
            the_fires[i].the_object.bg_color = YELLOW;
            the_fires[i].the_object.text_color = WHITE;
            strcpy(the_fires[i].the_object.symbol,"  ");
            
            the_fires[i].lifetime = lifetime;
            the_fires[i].delay = delay;
            the_fires[i].direction = direction;
            
            return;
        }
    }
}

void player_gets_hit()
{
    lives--;
    if (lives < 0) {
        set_cursor(28,1);
        set_bg_color(CYAN);
        set_text_color(WHITE);
        printf("Game Over");
    } else {
        load_level(level);
    }
}

void enemy_gets_hit(int x, int y)
{
    int i;
    for (i = 0; i < MAX_ENEMIES; i++)
    {
        if (the_enemies[i].alive > 0)
        {
            set_cursor(1,24);
            if (the_enemies[i].the_object.cur_x == x && the_enemies[i].the_object.cur_y == y) {
                clear_field(the_enemies[i].the_object.cur_x, the_enemies[i].the_object.cur_y);
                world[x][y] = 0;
                the_enemies[i].alive = 0;
            }
        }
    }
    
    int count = 0;
    for (i = 0; i < MAX_ENEMIES; i++)
    {
        if (the_enemies[i].alive > 0)
        {
            count++;
            break;
        }
    }
    
    if (count <= 0)
        level++;
        
}

void block_gets_hit(int x, int y)
{
    clear_field(x,y);
    world[x][y] = 0;
    int number = rand() % 8;
    set_cursor(1,24);
    if (number == 0) {
        if (the_player.max_bombs < 10)
            the_player.max_bombs++;
    } else if (number == 1) {
        if (the_player.range < 5)
            the_player.range++;
    }
    draw_status();
}

int load_level(unsigned int level)
{
    char filename[32];
    sprintf(filename, "level%d.txt", level);
    FILE *f = fopen(filename, "r");
    if (f == 0)
        return -1;
    
    int i;
    for (i = 0; i < MAX_ENEMIES; i++)
        the_enemies[i].alive = 0;
    for (i = 0; i < MAX_BOMBS; i++)
        the_bombs[i].countdown = 0;
    for (i = 0; i < MAX_FIRES; i++)
        the_fires[i].lifetime = 0;
    
    int x,y;
    for (y = 0; y < WORLD_HEIGHT; y++)
    {
        for (x = 0; x < WORLD_WIDTH; x++)
        {
            int tmp = fgetc(f) - 48;
            if (tmp == 2) { /* player */
                init_player(x, y);
            } else if (tmp == 3) { /* enemy */
                add_enemy(x,y);
            }
            world[x][y] = tmp;
        }
        /* ignore new line */
        fgetc(f);
    }
    
    fclose(f);
    
    draw_screen();
    draw_level();
    
    return 1;
}

void draw_level()
{
    int x,y;
    for (y = 0; y < WORLD_HEIGHT; y++)
    {
        for (x = 0; x < WORLD_WIDTH; x++)
        {
            set_cursor(2*x+1,y+1);
            if (world[x][y] == 1) {
                set_bg_color(WHITE);
                printf("  ");
            } else {
                set_bg_color(BLACK);
                printf("  ");
            }
        }
    }
    set_cursor(1,24);
    reset_colors();
}

void move_player(int input)
{
    if (input >= 0 && lives >= 0 && all_levels_completed <= 0) {
        switch(input) {
            case 'w': {
                int new_y = the_player.the_object.cur_y - 1;
                if((new_y) < 0) {
                    new_y = WORLD_HEIGHT - 1;
                }
                if (world[the_player.the_object.cur_x][new_y] == 0) {
                    the_player.the_object.old_y = the_player.the_object.cur_y;
                    the_player.the_object.cur_y = new_y;
                } else if (world[the_player.the_object.cur_x][new_y] == 3) {
                    player_gets_hit();
                }
            } break;
            case 'a': {
                int new_x = the_player.the_object.cur_x - 1;
                if((new_x) < 0) {
                    new_x = WORLD_WIDTH - 1;
                }
                if (world[new_x][the_player.the_object.cur_y] == 0) {
                    the_player.the_object.old_x = the_player.the_object.cur_x;
                    the_player.the_object.cur_x = new_x;
                } else if (world[new_x][the_player.the_object.cur_y] == 3) {
                    player_gets_hit();
                }
            } break;
            case 's': {
                int new_y = the_player.the_object.cur_y + 1;
                if((new_y) > WORLD_HEIGHT - 1) {
                    new_y = 0;
                }
                if (world[the_player.the_object.cur_x][new_y] == 0) {
                    the_player.the_object.old_y = the_player.the_object.cur_y;
                    the_player.the_object.cur_y = new_y;
                } else if (world[the_player.the_object.cur_x][new_y] == 3) {
                    player_gets_hit();
                }
            } break;
            case 'd': {
                int new_x = the_player.the_object.cur_x + 1;
                if((new_x) > WORLD_WIDTH - 1) {
                    new_x = 0;
                }
                if (world[new_x][the_player.the_object.cur_y] == 0) {
                    the_player.the_object.old_x = the_player.the_object.cur_x;
                    the_player.the_object.cur_x = new_x;
                } else if (world[new_x][the_player.the_object.cur_y] == 3) {
                    player_gets_hit();
                }
            } break;
            case ' ': {
                if (the_player.bombs < the_player.max_bombs) {
                    add_bomb(the_player.the_object.cur_x, the_player.the_object.cur_y);
                }
            } break;
            default:
                break;
        }
    }
    world[the_player.the_object.old_x][the_player.the_object.old_y] = 0;
    world[the_player.the_object.cur_x][the_player.the_object.cur_y] = 2;
    draw_object(&the_player.the_object);
}
    
void move_enemies()
{
    if (lives >= 0 && all_levels_completed <= 0) {
        int i;
        for (i = 0; i < MAX_ENEMIES; i++) {  
            if (the_enemies[i].alive) {
                int direction = rand() % 4;
                switch(direction) {
                    case 0: {
                        int new_y = the_enemies[i].the_object.cur_y - 1;
                        if((new_y) < 0) {
                            new_y = WORLD_HEIGHT - 1;
                        }
                        if (world[the_enemies[i].the_object.cur_x][new_y] == 0) {
                            the_enemies[i].the_object.old_y = the_enemies[i].the_object.cur_y;
                            the_enemies[i].the_object.cur_y = new_y;
                        } else if (world[the_enemies[i].the_object.cur_x][new_y] == 2) {
                            player_gets_hit();
                        }
                    } break;
                    case 1: {
                        int new_x = the_enemies[i].the_object.cur_x - 1;
                        if((new_x) < 0) {
                            new_x = WORLD_WIDTH - 1;
                        }
                        if (world[new_x][the_enemies[i].the_object.cur_y] == 0) {
                            the_enemies[i].the_object.old_x = the_enemies[i].the_object.cur_x;
                            the_enemies[i].the_object.cur_x = new_x;
                        } else if (world[new_x][the_enemies[i].the_object.cur_y] == 2) {
                            player_gets_hit();
                        }
                    } break;
                    case 2: {
                        int new_y = the_enemies[i].the_object.cur_y + 1;
                        if((new_y) > WORLD_HEIGHT - 1) {
                            new_y = 0;
                        }
                        if (world[the_enemies[i].the_object.cur_x][new_y] == 0) {
                            the_enemies[i].the_object.old_y = the_enemies[i].the_object.cur_y;
                            the_enemies[i].the_object.cur_y = new_y;
                        } else if (world[the_enemies[i].the_object.cur_x][new_y] == 2) {
                            player_gets_hit();
                        }
                    } break;
                    case 3: {
                        int new_x = the_enemies[i].the_object.cur_x + 1;
                        if((new_x) > WORLD_WIDTH - 1) {
                            new_x = 0;
                        }
                        if (world[new_x][the_enemies[i].the_object.cur_y] == 0) {
                            the_enemies[i].the_object.old_x = the_enemies[i].the_object.cur_x;
                            the_enemies[i].the_object.cur_x = new_x;
                        } else if (world[new_x][the_enemies[i].the_object.cur_y] == 2) {
                            player_gets_hit();
                        }
                    } break;
                    default:
                        break;
                }
                world[the_enemies[i].the_object.old_x][the_enemies[i].the_object.old_y] = 0;
                world[the_enemies[i].the_object.cur_x][the_enemies[i].the_object.cur_y] = 3;
                draw_object(&the_enemies[i].the_object);
            }
        }
    }
}

void move_bombs()
{
    if (lives >= 0 && all_levels_completed <= 0) {
        int i;
        for (i = 0; i < MAX_BOMBS; i++) {  
            if (the_bombs[i].countdown > 0) {
                the_bombs[i].countdown--;
                if (the_bombs[i].countdown <= 0)
                {
                    clear_field(the_bombs[i].the_object.cur_x, the_bombs[i].the_object.cur_y);
                    //world[the_bombs[i].the_object.cur_x][the_bombs[i].the_object.cur_y] = 0;
                    add_fire(the_bombs[i].the_object.cur_x,
                             the_bombs[i].the_object.cur_y,
                             the_bombs[i].range+1,0,0);
                    add_fire(the_bombs[i].the_object.cur_x,
                             the_bombs[i].the_object.cur_y,
                             the_bombs[i].range+1,0,1);
                    add_fire(the_bombs[i].the_object.cur_x,
                             the_bombs[i].the_object.cur_y,
                             the_bombs[i].range+1,0,2);
                    add_fire(the_bombs[i].the_object.cur_x,
                             the_bombs[i].the_object.cur_y,
                             the_bombs[i].range+1,0,3);
                    the_player.bombs--;
                } else {
                    char label[3];
                    sprintf(label, "%2i", the_bombs[i].countdown);
                    strcpy(the_bombs[i].the_object.symbol,label);
                    draw_object(&the_bombs[i].the_object);
                }
            }
        }
    }
}

void move_fires()
{
    if (lives >= 0 && all_levels_completed <= 0) {
        int i;
        for (i = 0; i < MAX_FIRES; i++) {  
            if (the_fires[i].lifetime > 0) {
                the_fires[i].lifetime--;
                
                if (world[the_fires[i].the_object.cur_x][the_fires[i].the_object.cur_y] == 2) {
                    player_gets_hit();
                } else if (world[the_fires[i].the_object.cur_x][the_fires[i].the_object.cur_y] == 3) {
                    enemy_gets_hit(the_fires[i].the_object.cur_x,the_fires[i].the_object.cur_y);
                }
                
                if (the_fires[i].lifetime > 0) {
                    switch(the_fires[i].direction) {
                        case 0: {
                            int new_y = the_fires[i].the_object.cur_y - 1;
                            if((new_y) < 0) {
                                new_y = WORLD_HEIGHT - 1;
                            }
                            if (world[the_fires[i].the_object.cur_x][new_y] == 0) {
                                the_fires[i].the_object.old_y = the_fires[i].the_object.cur_y;
                                the_fires[i].the_object.cur_y = new_y;
                            } else if (world[the_fires[i].the_object.cur_x][new_y] == 2) {
                                player_gets_hit();
                            } else if (world[the_fires[i].the_object.cur_x][new_y] == 3) {
                                enemy_gets_hit(the_fires[i].the_object.cur_x,new_y);
                            }  else if (world[the_fires[i].the_object.cur_x][new_y] == 1) {
                                block_gets_hit(the_fires[i].the_object.cur_x,new_y);
                                the_fires[i].lifetime = 0;
                            }
                        } break;
                        case 1: {
                            int new_x = the_fires[i].the_object.cur_x - 1;
                            if((new_x) < 0) {
                                new_x = WORLD_WIDTH - 1;
                            }
                            if (world[new_x][the_fires[i].the_object.cur_y] == 0) {
                                the_fires[i].the_object.old_x = the_fires[i].the_object.cur_x;
                                the_fires[i].the_object.cur_x = new_x;
                            } else if (world[new_x][the_fires[i].the_object.cur_y] == 2) {
                                player_gets_hit();
                            } else if (world[new_x][the_fires[i].the_object.cur_y] == 3) {
                                enemy_gets_hit(new_x,the_fires[i].the_object.cur_y);
                            } else if (world[new_x][the_fires[i].the_object.cur_y] == 1) {
                                block_gets_hit(new_x,the_fires[i].the_object.cur_y);
                                the_fires[i].lifetime = 0;
                            }
                        } break;
                        case 2: {
                            int new_y = the_fires[i].the_object.cur_y + 1;
                            if((new_y) > WORLD_HEIGHT - 1) {
                                new_y = 0;
                            }
                            if (world[the_fires[i].the_object.cur_x][new_y] == 0) {
                                the_fires[i].the_object.old_y = the_fires[i].the_object.cur_y;
                                the_fires[i].the_object.cur_y = new_y;
                            } else if (world[the_fires[i].the_object.cur_x][new_y] == 2) {
                                player_gets_hit();
                            } else if (world[the_fires[i].the_object.cur_x][new_y] == 3) {
                                enemy_gets_hit(the_fires[i].the_object.cur_x,new_y);
                            } else if (world[the_fires[i].the_object.cur_x][new_y] == 1) {
                                block_gets_hit(the_fires[i].the_object.cur_x,new_y);
                                the_fires[i].lifetime = 0;
                            }
                        } break;
                        case 3: {
                            int new_x = the_fires[i].the_object.cur_x + 1;
                            if((new_x) > WORLD_WIDTH - 1) {
                                new_x = 0;
                            }
                            if (world[new_x][the_fires[i].the_object.cur_y] == 0) {
                                the_fires[i].the_object.old_x = the_fires[i].the_object.cur_x;
                                the_fires[i].the_object.cur_x = new_x;
                            } else if (world[new_x][the_fires[i].the_object.cur_y] == 2) {
                                player_gets_hit();
                            } else if (world[new_x][the_fires[i].the_object.cur_y] == 3) {
                                enemy_gets_hit(new_x,the_fires[i].the_object.cur_y);
                            } else if (world[new_x][the_fires[i].the_object.cur_y] == 1) {
                                block_gets_hit(new_x,the_fires[i].the_object.cur_y);
                                the_fires[i].lifetime = 0;
                            }
                        } break;
                        default:
                            break;
                    }
                    draw_object(&the_fires[i].the_object);
                }
                
                if (the_fires[i].lifetime <= 0)
                    clear_field(the_fires[i].the_object.old_x, the_fires[i].the_object.old_y);
            }
        }
    }
}

void draw_screen()
{
    clear_screen();
    
    set_cursor(1,20);
    set_bg_color(BLUE);
    set_text_color(WHITE);
	int i,j;
    for (i = 1; i < 25; i++) {
		set_cursor(61,i);
		printf("|");
	}
    
    set_bg_color(CYAN);
    set_text_color(WHITE);
    for (i = 62; i < 81; i++) {
        for(j = 1; j < 25; j++) {
            set_cursor(i,j);
            printf(" ");
        }
    }
	set_cursor(63,1);
	printf("W:     Up");
	set_cursor(63,2);
	printf("A:     Left");
	set_cursor(63,3);
	printf("S:     Down");
	set_cursor(63,4);
	printf("D:     Right");
    set_cursor(63,5);
	printf("Space: Place Bomb");
    
    set_cursor(63,7);
	printf("N:     New Game");
    set_cursor(63,8);
	printf("Q:     Quit");
    
    draw_status();
    
    reset_colors();
}

void draw_status()
{
    set_bg_color(CYAN);
    set_text_color(WHITE);
    set_cursor(63,19);
	printf("Level: %2i", level);
    
    set_cursor(63,21);
	printf("Lives: %2i", lives);
    set_cursor(63,22);
	printf("Range: %2i", the_player.range);
    set_cursor(63,23);
	printf("Bombs: %2i", the_player.max_bombs);
    reset_colors();
}

int getch()
{
	static int ch = -1, fd = 0;
	struct termios new, old;
	fd = fileno(stdin);
	tcgetattr(fd, &old);
	new = old;
	new.c_lflag &= ~(ICANON|ECHO);
	tcsetattr(fd, TCSANOW, &new);
	ch = getchar();
	tcsetattr(fd, TCSANOW, &old);
	return ch;
}

/* Drawing functions */

void draw_object(struct object* the_object)
{
    set_bg_color(BLACK);
	set_cursor(the_object->old_x * 2 + 1, the_object->old_y + 1);
	printf("  ");
	set_cursor(the_object->cur_x * 2 + 1, the_object->cur_y + 1);
    set_bg_color(the_object->bg_color);
    set_text_color(the_object->text_color);
	printf("%s",the_object->symbol);
    reset_colors();
	set_cursor(1, 24);
	fflush(stdout);
	the_object->old_x = the_object->cur_x;
	the_object->old_y = the_object->cur_y;
}

void clear_field(int x, int y)
{
    set_bg_color(BLACK);
    set_cursor(x * 2 + 1, y + 1);
    printf("  ");
    set_cursor(1, 24);
}

void clear_screen()
{
	printf("\e[2J");
}

void set_cursor(int x, int y)
{
	printf("\e[%d;%dH",y,x);
}

void set_text_color(enum color the_color)
{
    printf("\e[3%dm", the_color);
}

void set_bg_color(enum color the_color)
{
    printf("\e[4%dm", the_color);
}

void reset_colors()
{
    printf("\e[0m");
}
